#---------------------------------------------------------------------------------------
# Some of the code snippets are from 
# http://www.pierssen.com/arcgis10/upload/python/snippets102.py
#
# more details on arcobject/com using pyghon and links @
# https://gis.stackexchange.com/questions/80/accessing-arcobjects-from-python
#---------------------------------------------------------------------------------------

from __future__ import print_function
from collections import namedtuple
import xml.etree.ElementTree as et
import winreg
import os
import time
from random import randrange

try:
    import comtypes
except ImportError:
    import sys
    import subprocess
    
    while True:
        response = raw_input("""
            'comtypes' module not found. Proceed with installation (y/n)? """)
        if response == 'Y' or response == 'y':
            subprocess.check_call([sys.executable, "-m", "pip", "install", "comtypes"])
            print("\n\n")
            import comtypes
            break
        elif response == 'N' or response == 'n':
            sys.exit("""
            Requires installation of comtypes module! (https://pypi.org/project/comtypes/)
            run the following command to install: 
            python -m pip install comtypes.
            """)
            break
        else:
            print("""
            Response should be one of the expected responses: y, n""")

def GetProInstallDir_AutoDetect():
    try:
        keyPro = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\ESRI\\ArcGISPro")
        return winreg.QueryValueEx(keyPro, "InstallDir")[0]
    except:
        return ''

def GetArcMapInstallDir_AutoDetect():
    try:
        from comtypes.client import GetModule
        keyEsriArcGIS = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\WOW6432Node\\Esri\\ArcGIS")
        version = winreg.QueryValueEx(keyEsriArcGIS, "RealVersion")[0]
        maj, min, _ = version.split('.')
        version = '.'.join([maj, min])
        arcmapDesktopRegPath = os.path.join("SOFTWARE\\WOW6432Node\\Esri", "Desktop" + version)
        keyArcmapDesktop = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, arcmapDesktopRegPath)
        arcmapInstallDir = winreg.QueryValueEx(keyArcmapDesktop, "InstallDir")[0]
        return arcmapInstallDir
    except:
        return ''

def GetProductionMappingInstallDir_AutoDetect():
    try:
        from comtypes.client import GetModule
        keyEsriProd = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\WOW6432Node\\EsriProduction")
        version = winreg.QueryValueEx(keyEsriProd, "RealVersion")[0]
        maj, min, _ = version.split('.')
        version = '.'.join([maj, min])
        esriProdMappRegPath = os.path.join("SOFTWARE\\WOW6432Node\\EsriProduction", "Desktop" + version, "EsriProductionMapping")
        keyEsriProdMapp = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, esriProdMappRegPath)
        prodMappPath = winreg.QueryValueEx(keyEsriProdMapp, "ProductionMappingDir")[0]
        return prodMappPath
    except:
        return ''

def SetArcMapInstallDir(path):
    global arcmap_dir
    arcmap_dir = path

def GetArcMapInstallDir():
    global arcmap_dir
    return arcmap_dir

def SetProductionMappingInstallDir(path):
    global production_mapping_dir
    production_mapping_dir = path

def GetProductionMappingInstallDir():
    global production_mapping_dir
    return production_mapping_dir

def GetLibPath():
    return os.path.join(GetArcMapInstallDir(),  "com")

def GetModule(sModuleName):
    """Import ArcGIS module"""
    from comtypes.client import GetModule
    sLibPath = GetLibPath()
    return GetModule(os.path.join(sLibPath, sModuleName))

def GetProductionMappingModule(sModuleName):
    prodMappBinPath = os.path.join(GetProductionMappingInstallDir(),  'Bin')
    return comtypes.client.GetModule(os.path.join(prodMappBinPath, sModuleName))

def NewObj(MyClass, MyInterface):
    """Creates a new comtypes POINTER object where\n\
    MyClass is the class to be instantiated,\n\
    MyInterface is the interface to be assigned"""
    from comtypes.client import CreateObject
    ptr = CreateObject(MyClass, interface=MyInterface)
    return ptr

def InitLicense():
    """Init standalone ArcGIS license"""
    # Set ArcObjects version
    esriSystem = GetModule("esriSystem.olb")
    # Get license
    pInit = NewObj(esriSystem.AoInitialize, esriSystem.IAoInitialize)
    ProductList = [esriSystem.esriLicenseProductCodeAdvanced, \
                   esriSystem.esriLicenseProductCodeStandard, \
                   esriSystem.esriLicenseProductCodeBasic]
    for eProduct in ProductList:
        licenseStatus = pInit.IsProductCodeAvailable(eProduct)
        if licenseStatus != esriSystem.esriLicenseAvailable:
            continue
        licenseStatus = pInit.Initialize(eProduct)
        return (licenseStatus == esriSystem.esriLicenseCheckedOut)
    return False

def OpenWorkspace(path):
    pWorkspaceFactory = None
    esriDataSourcesGDB =  GetModule("esriDataSourcesGDB.olb")
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    if (path.endswith('.sde')):
        pWorkspaceFactory = NewObj(esriDataSourcesGDB.SdeWorkspaceFactory, esriGeoDatabase.IWorkspaceFactory)
    elif (path.endswith('.gdb')):
        pWorkspaceFactory = NewObj(esriDataSourcesGDB.FileGDBWorkspaceFactory, esriGeoDatabase.IWorkspaceFactory)
    pWorkspace = pWorkspaceFactory.OpenFromFile(path, 0)
    return pWorkspace
   
def OpenFeatureWorkspace(path):
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    pWorkspace = OpenWorkspace(path)
    pFeatureWorkspace =  pWorkspace.QueryInterface(esriGeoDatabase.IFeatureWorkspace)
    return pFeatureWorkspace

def OpenTable(workspacePath, tableName):
    pFeatureWorkspace =  OpenFeatureWorkspace(workspacePath)
    return pFeatureWorkspace.OpenTable(tableName)
    # pFeatureClass =  pFeatureWorkspace.OpenFeatureClass(featureClassName)

def OpenRelationshipClass(workspacePath, relnClassName):
    pFeatureWorkspace = OpenFeatureWorkspace(workspacePath)
    return pFeatureWorkspace.OpenRelationshipClass(relnClassName)

def StartEditSession(workspacePath, versioned=True):
    pWorkspace  = OpenWorkspace(workspacePath)
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    pWkspEdit =  pWorkspace.QueryInterface(esriGeoDatabase.IWorkspaceEdit2)
    if pWkspEdit.IsBeingEdited():
        raise Exception("Workspace is currently being edited in another context!")
    if (workspacePath.endswith('.sde')):
        pMUWkspEdit = pWorkspace.QueryInterface(esriGeoDatabase.IMultiuserWorkspaceEdit)
        if versioned:
            pMUWkspEdit.StartMultiuserEditing(esriGeoDatabase.esriMESMVersioned)
        else: # use this for editing non-versioned tables/feature classes
            pMUWkspEdit.StartMultiuserEditing(esriGeoDatabase.esriMESMNonVersioned)
    elif (workspacePath.endswith('.gdb')):
        pWkspEdit.StartEditing(False)

def StopEditSession(workspacePath):
    pWorkspace  = OpenWorkspace(workspacePath)
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    pWkspEdit =  pWorkspace.QueryInterface(esriGeoDatabase.IWorkspaceEdit2)
    pWkspEdit.StopEditing(True)

def StartEditOperation(workspacePath):
    pWorkspace  = OpenWorkspace(workspacePath)
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    pWkspEdit =  pWorkspace.QueryInterface(esriGeoDatabase.IWorkspaceEdit2)
    if pWkspEdit.IsInEditOperation:
        raise Exception("Worksapce has pending edit operations!")
    pWkspEdit.StartEditOperation()

def StopEditOperation(workspacePath):
    pWorkspace  = OpenWorkspace(workspacePath)
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    pWkspEdit =  pWorkspace.QueryInterface(esriGeoDatabase.IWorkspaceEdit2)
    pWkspEdit.StopEditOperation()

# creates row with field/values dictionary, insert into the table, and returns the row
def InsertIntoTable(pTable, fieldValuesDict):
    pRow = pTable.CreateRow()
    for key in fieldValuesDict.keys():
        iField = pTable.FindField(key)
        val = fieldValuesDict[key]
        if val is None:
            val = pTable.Fields.Field(iField).DefaultValue
        pRow.Value[iField] = val
    pRow.Store()
    return pRow

# returns rows from one or more table(s)
def GetTablesRows(workspacePath, tables, whereClause = '', subfields = '*'):
    pFeatureWorkspace = OpenFeatureWorkspace(workspacePath)
    queryDef = pFeatureWorkspace.CreateQueryDef()
    queryDef.Tables = tables
    queryDef.SubFields = subfields
    queryDef.WhereClause = whereClause
    pCursor = queryDef.Evaluate() # QueryDef cursors are readonly.
    rows = []
    pRow = pCursor.NextRow()
    while pRow:
        rows.append(pRow)
        pRow = pCursor.NextRow()
    return rows

# convert IRow to a named tuple of columns
# xml field value is parsed as xml.etree.ElementTree
# blob is converted to IMemoryBlobStream
def RowToNamedTuple(pRow, fieldAlias=''):
    fieldnames = []
    if fieldAlias != '':
        fieldnames = [x.strip() for x in fieldAlias.split(',')]
        if(len(fieldnames) != pRow.Fields.FieldCount):
            raise Exception('Alias list should be comma delimited and should match the number of fields!')
    else:
        for i in range(0, pRow.Fields.FieldCount):
            fieldnames.append(pRow.Fields.Field(i).Name.upper())
    FieldsTuple = namedtuple('Fields', fieldnames)

    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    vals = []
    for i in range(0, pRow.Fields.FieldCount):
        val = pRow.Value(i)
        if pRow.Fields.Field(i).type == esriGeoDatabase.esriFieldTypeXML:
            strval = ''.join([chr(x) for x in val]) 
            val = et.fromstring(strval)
        elif pRow.Fields.Field(i).type == esriGeoDatabase.esriFieldTypeBlob:
            esriSystem = GetModule("esriSystem.olb")
            val = val.QueryInterface(esriSystem.IMemoryBlobStream)
        vals.append(val)
    return FieldsTuple(*vals)

def RegisterMaritimeClassExtension(pTable):
    esriSystem = GetModule("esriSystem.olb")
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    pUID = NewObj(esriSystem.UID, esriSystem.IUID)
    pUID.Value = "{27FA073D-1BF8-4C8A-8752-2D32403928CD}" # Maritime Class extension
    pIObjClass =  pTable.QueryInterface(esriGeoDatabase.IObjectClass)
    pSchemaLock = pIObjClass.QueryInterface(esriGeoDatabase.ISchemaLock)
    pSchemaLock.ChangeSchemaLock(esriGeoDatabase.esriExclusiveSchemaLock)
    pSchemaEdit = pIObjClass.QueryInterface(esriGeoDatabase.IClassSchemaEdit2)
    pSchemaEdit.AlterClassExtensionCLSID(pUID, None)
    pSchemaLock.ChangeSchemaLock(esriGeoDatabase.esriSharedSchemaLock)


def UnRegisterMaritimeClassExtension(pTable):
    esriSystem = GetModule("esriSystem.olb")
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")

    pIObjClass =  pTable.QueryInterface(esriGeoDatabase.IObjectClass)
    pSchemaLock = pIObjClass.QueryInterface(esriGeoDatabase.ISchemaLock)
    pSchemaLock.ChangeSchemaLock(esriGeoDatabase.esriExclusiveSchemaLock)
    pSchemaEdit = pIObjClass.QueryInterface(esriGeoDatabase.IClassSchemaEdit2)
    pSchemaEdit.AlterClassExtensionCLSID(None, None)
    pSchemaLock.ChangeSchemaLock(esriGeoDatabase.esriSharedSchemaLock)

def RegisterMaritimeWorkspaceExtension(pWksp):
    esriSystem = GetModule("esriSystem.olb")
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    pUID = NewObj(esriSystem.UID, esriSystem.IUID)
    pUID.Value = "{ec3210c0-0c2e-45ed-a708-5714098736d8}"
    pWkspExtManager = pWksp.QueryInterface(esriGeoDatabase.IWorkspaceExtensionManager)

    # unregister first, TODO: try to make pWkspExtManager.FindExtension(pUID) work instead of unregistering first 
    pWkspExtManager._IWorkspaceExtensionManager__com_UnRegisterExtension(pUID)
    try:
        pWkspExtManager.RegisterExtension("MaritimeWorkspaceExtension", pUID)
    except Exception as ex:
        print('Failed to register workspace extension. {}'.format(ex.message))
        # if failed try running register_wksp_ext.py script with admin connection.

def GetPLTSSQLString(ipMemoryBlobStream):
    prodSolnEngine = GetProductionMappingModule("ProductionSolutionEngine.olb")
    esriSystem = GetModule("esriSystem.olb")
    pStream = ipMemoryBlobStream.QueryInterface(esriSystem.IStream)
    pltsSqlStmt = NewObj(prodSolnEngine.PLTSSQLStatement, prodSolnEngine.IPLTSSQLStatement)
    pPStream = pltsSqlStmt.QueryInterface(esriSystem.IPersistStream)
    pPStream.Load(pStream)
    qry = pltsSqlStmt.SQLString
    return qry

# method 1. 
# Queries the FREL table for all NOIDs, for each Feature class update the NOIDs to GUIDs,
# for each feature class, update the dictionary created in first step for all participating NOIDS
# Updates the FREL table after all the feature classes have been updated with the dict updated in 2nd step.
def CopyGuidToNoid(workspacePath, featureClassList, working_dir, indent = '\t\t\t\t'):
    
    # Save all the guid to noid (that was replaced) in a file.
    gdbname = os.path.split(workspacePath)[1].replace('.', '_')
    noidDumpFilename = os.path.join(working_dir, 'GuidToNoid_' + gdbname + "__0" )
    while(os.path.exists(noidDumpFilename + ".txt")):
        prefix, postfix = noidDumpFilename.split('__')
        count = int(postfix) + 1
        noidDumpFilename = prefix + '__' + str(count)
    noidDump = open(noidDumpFilename + '.txt', 'w+')

    # step 1 Create a dict of all NOIDs participating in relationship. Leave the Guid blank.
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    pQF = NewObj(esriGeoDatabase.QueryFilter, esriGeoDatabase.IQueryFilter)
    pFRELTable = OpenTable(workspacePath, 'PLTS_FREL')
    pQF.SubFields = "SRC_UID, DEST_UID"
    pCursor = pFRELTable.Search(pQF, True) # use recycling cursor for performance
    # The query filter returns all fields, but only those specified by the SubFields are populated.
    # So we still need to find the index of the field we are looking for.
    srcIndex = pCursor.FindField("SRC_UID")
    destIndex = pCursor.FindField("DEST_UID")
    pRow = pCursor.NextRow()
    dictNoidsToGuids = {}
    while pRow:
        srcNoid = pRow.Value[srcIndex]
        if srcNoid is not None and srcNoid.strip() != '':
            dictNoidsToGuids[srcNoid] = ''
        destNoid = pRow.Value[destIndex]
        if destNoid is not None and destNoid.strip() != '':
            dictNoidsToGuids[destNoid] = ''
        pRow = pCursor.NextRow()

    # step 2. Replace the GUIDS in all of the feature classes, and Populate the Guids in the dict for the features 
    # participating in the relationshps
    for fc in featureClassList:
        pQF.SubFields = "GlobalId, NOID"
        pTable = OpenTable(workspacePath, fc)
        pCursor = pTable.Search(pQF, False)
        # The query filter returns all fields, but only those specified by the SubFields are populated.
        # So we still need to find the index of the field we are looking for.
        guidIndex = pCursor.FindField("GlobalId")
        noidIndex = pCursor.FindField("NOID")
        if (guidIndex <= 0 or noidIndex <= 0):
            continue
        # not used in relationships, so just skip it (typically has too many rows to process in one session)
        if fc == 'SoundingsP':
            continue
        print(indent + ' ' + fc, end = ' ... ')
        pRow = pCursor.NextRow()
        while pRow:
            noid = pRow.Value[noidIndex]
            guid = pRow.Value[guidIndex]
            if noid is None:
                noid = '<NULL>'
            noidDump.write(fc + ',' + guid + ',' + noid + '\n')
            pRow.Value[noidIndex] = guid
            pRow.Store()
            if noid in dictNoidsToGuids:
                dictNoidsToGuids[noid] = guid
            pRow = pCursor.NextRow()
        print ('OK')

    # step 3. Update the FREL table using the dictionary of NOIDs to GUIDs
    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    pQF.SubFields = "SRC_UID, DEST_UID"
    pCursor = pFRELTable.Search(pQF, False)
    pRow = pCursor.NextRow()
    while pRow:
        srcNoid = pRow.Value[srcIndex]
        destNoid = pRow.Value[destIndex]
        if srcNoid in dictNoidsToGuids:
            pRow.Value[srcIndex] = dictNoidsToGuids[srcNoid]
            pRow.Store()
        if destNoid in dictNoidsToGuids:
            pRow.Value[destIndex] = dictNoidsToGuids[destNoid]
            pRow.Store()
        pRow = pCursor.NextRow()


# method 2.
# For each feature classes, updates NOID to GUID, and updates the PFREL corresponding to each Feature Classes.
# This method queries PFREL table for each feature class.
# The first method queries PFREL once at beginning and once at the end and carried out in 1 edit session.
# method 1 is little bit faster. However due to the oracle issue:
# "ORA-30036: unable to extend segment by 8 in undo tablespace 'UNDOTBS'",
# we will have to split a huge transaction into per feature class transaction. 
# So imperative to update the related NOIDs in the FREL table as we process the src or dest feature classes, 
# so that in cases when the migration is done in multiple runs, the NOIDs in a FC as well as the 
# related NOIDs in the FREL table are updated in one session(transaction).
def CopyGUIDToNoid2(workspacePath, featureClassList, working_dir, chunkSize, verbose, indent = '\t\t\t\t'):

    import arcpy

    # Save all the guid to noid (that was replaced) in a file.
    gdbname = os.path.split(workspacePath)[1].replace('.', '_')
    noidBackupFilename = os.path.join(working_dir, gdbname + '_FREL_GUID2NOID_backup.txt')
    noidBackupFile = open(noidBackupFilename, 'a+')
    guid2NoidBackup = dict()
    for line in noidBackupFile.readlines():
        gn = line.replace('\n', '').split(',')
        if len(gn) > 2:
            guid2NoidBackup[gn[1]] = gn[2]

    # save already migrated fcs, so on repeated runs these are skipped.
    migratedFCs = dict()
    migratedFCsFile = open(os.path.join(working_dir,  gdbname + '_MigratedFeatureClasses.txt'), 'a+')
    for line in migratedFCsFile.readlines():
        fc_oid = line.replace('\n', '').split(':')
        fc = fc_oid[0].strip()
        migratedFCs[fc] = ''
        if len(fc_oid) > 1:
            migratedFCs[fc] = fc_oid[1].strip()
    migratedFCsFile.seek(0)
    migratedFCsFile.truncate()

    esriGeoDatabase = GetModule("esriGeoDatabase.olb")
    pQF = NewObj(esriGeoDatabase.QueryFilter, esriGeoDatabase.IQueryFilter)

    totalRowsCommittedAfterLastCompress = 0

    for fc in featureClassList:
        nProcessedRows = 0
        if fc in migratedFCs:
            completed = False
            if(migratedFCs[fc] == '' or 'complete' in migratedFCs[fc]): 
                completed = True
            else:
                nr = migratedFCs[fc].split('/')
                nProcessedRows = int(nr[0])
                total = int(nr[1])
                if(nProcessedRows == total):
                    completed = True
            
            if completed:
                print(indent + ' Skipping migrated FeatureClass ' + fc + ' ... ')
                migratedFCsFile.write(fc + ' : ' + migratedFCs[fc] + '\n')
                migratedFCsFile.flush()
                continue
        
        # not used in relationships, so just skip it (typically has too many rows to process in one session)
        if fc == 'SoundingsP':
            continue

        pTableInit = OpenTable(workspacePath, fc)
        # The query filter returns all fields, but only those specified by the SubFields are populated.
        # So we still need to find the index of the field we are looking for.
        guidIndex = pTableInit.FindField("GlobalId")
        noidIndex = pTableInit.FindField("NOID")
        oidFieldName = pTableInit.OIDFieldName
        oidIndex = pTableInit.FindField(oidFieldName)
        if (guidIndex < 0 or noidIndex < 0 or oidIndex < 0):
            continue

        pQF.SubFields = oidFieldName
        pQF.WhereClause = ''
        pCursorInit = pTableInit.Search(pQF, False)
        pRowInit = pCursorInit.NextRow()
        listOids = []
        while pRowInit:
            listOids.append(pRowInit.Value[oidIndex])
            pRowInit = pCursorInit.NextRow()
        listOids.sort()

        noid_msg_prefix =  indent + ' ' + fc + ' ... '
        if (nProcessedRows > 0):
            print(noid_msg_prefix + 'Continuing from last successfully updated NOID ... ')
        else:
            print(noid_msg_prefix)
        
        totalRows = len(listOids)
        for index in range(nProcessedRows, totalRows, chunkSize):
            lowerBound = listOids[index]
            upperBoundIndex = (index + chunkSize - 1) if (index + chunkSize - 1) < totalRows else (totalRows-1)
            upperBound = listOids[upperBoundIndex]

            if verbose:
                print(indent + "\t starting edit session")
            StartEditSession(workspacePath)
            StartEditOperation(workspacePath)

            if verbose:
                print(indent + "\t opening feature class and querying next {} rows".format(chunkSize))
            pTableUpdate = OpenTable(workspacePath, fc)
            pQF.SubFields = "GlobalId, NOID"
            pQF.WhereClause = oidFieldName + " >= " + str(lowerBound) + " AND " + oidFieldName + " <= " + str(upperBound)
            pCursorUpdate = pTableUpdate.Search(pQF, False)

            noidToGuid = {}
            setOfNoidsInFREL = set()

            if verbose:
                print(indent + "\t replacing noid in the returned rows")

            count_processed_row  = 0
            pRowUpdate = pCursorUpdate.NextRow()
            while pRowUpdate:
                noid = pRowUpdate.Value[noidIndex]
                guid = pRowUpdate.Value[guidIndex]
                if noid is not None and noid.strip() != '':
                    noidToGuid[noid] = guid
                else:
                    noid = '<NULL> or Invalid NOID'
                pRowUpdate.Value[noidIndex] = guid
                pRowUpdate.Store()
                count_processed_row += 1
                if verbose:
                    if(count_processed_row % 1000 == 0):
                        print(indent + "\t\t updated {} rows".format(count_processed_row))
                pRowUpdate = pCursorUpdate.NextRow()

            if verbose:
                print(indent + "\t updating plts frel table")
            # This method queries the PLTS_FREL for each feature class for each chunk.
            # update the NOIDs in PLTS_FREL table
            # all tables must be opened inside edit session to actually save the change
            pFRELTable = OpenTable(workspacePath, 'PLTS_FREL')
            pQF.SubFields = "SRC_UID, DEST_UID, SRC_FC"
            pQF.WhereClause = "SRC_FC = '"+ fc +"' OR DEST_FC = '" + fc + "'"
            pCursorFREL = pFRELTable.Search(pQF, False)
            # The query filter returns all fields, but only those specified by the SubFields are populated.
            # So we still need to find the index of the field we are looking for.
            srcIndex = pCursorFREL.FindField("SRC_UID")
            destIndex = pCursorFREL.FindField("DEST_UID")
            srcFCIndex = pCursorFREL.FindField("SRC_FC")
            pRowFREL = pCursorFREL.NextRow()
            while pRowFREL:
                srcNoid = pRowFREL.Value[srcIndex]
                destNoid = pRowFREL.Value[destIndex]
                srcFC = pRowFREL.Value[srcFCIndex]
                if srcFC == fc:
                    if srcNoid in noidToGuid:
                        pRowFREL.Value[srcIndex] = noidToGuid[srcNoid]
                        setOfNoidsInFREL.add(srcNoid)
                        pRowFREL.Store()  
                else: # destFC == fc
                    if destNoid in noidToGuid:
                        pRowFREL.Value[destIndex] = noidToGuid[destNoid]
                        setOfNoidsInFREL.add(destNoid)
                        pRowFREL.Store()
                pRowFREL = pCursorFREL.NextRow()

            if verbose:
                print(indent + "\t stopping edit session")
            StopEditOperation(workspacePath)
            StopEditSession(workspacePath)
            for noid in setOfNoidsInFREL:
                if (noid not in guid2NoidBackup):
                    noidBackupFile.write(fc + ',' + noidToGuid[noid] + ',' + noid + '\n')
            noidBackupFile.flush()

            migratedFCsFile.write('{} : {}/{}\n'.format(fc, upperBoundIndex + 1, totalRows))
            migratedFCsFile.flush()
            # end = '\n' if upperBoundIndex >= totalRows -1 else '\r'
            print('{} {}/{} committed \n'.format(noid_msg_prefix, upperBoundIndex + 1, totalRows))
    
        totalRowsCommittedAfterLastCompress += totalRows

        # compress the database after each feature class is updated. 
        # This is required for versioned sde, since each edits are tracked in the delta tables and grows in size and
        # make sure it is an admin connection.
        desc = arcpy.Describe(workspacePath)
        if(desc.workspaceType == 'RemoteDatabase'):
            if(totalRowsCommittedAfterLastCompress >= 500000):
                cmp_start_msg = indent + "\t Compressing database ... "
                print(cmp_start_msg, end='\r')
                try:
                    arcpy.Compress_management(workspacePath)
                    print(cmp_start_msg +  " complete")
                    totalRowsCommittedAfterLastCompress = 0
                except:
                    print(cmp_start_msg  + " Unable to compress. Continuing without compressing ... ", end="\n")


    migratedFCsFile.close()
    noidBackupFile.close()

def GenerateFIDN():
    return '{:010d}'.format(randrange(1, 0xFFFFFFFF - 1))
