from __future__ import print_function
import os
import arcpy
import operator
from com_utils import *

from collections import namedtuple
import xml.etree.ElementTree as et

import time
import sys
import datetime
import copy

# The logging module also have side-effect of logs form other modules so using a custom Logger class
class Logger(object):
    def __init__(self, logdir, name_prefix):
        self.terminal = sys.stdout
        self.log = open(os.path.join(logdir , 'log_' + name_prefix + '_' + str(int(time.time()*10))) + '.txt', 'w')

    def write(self, message):
        if(message.strip() != ''): #if it is a whitespace(for example \n) do not append the time 
            message = str(datetime.datetime.now()) + ' - ' + message
        self.terminal.write(message.encode('utf8'))
        self.log.write(message.encode('utf8')) 

def Migrate(arcmap_dir, production_mapping_dir, pl, nis_paths_orig, working_dir, chunkSize, agency, fids, verbose):
    gdbname = os.path.split(pl)[1].replace('.', '_')
    logger = Logger(working_dir, gdbname)
    sys.stdout = logger
    sys.stderr = logger

    if(arcmap_dir.strip() == ''):
        arcmap_dir = GetArcMapInstallDir_AutoDetect()
        if(arcmap_dir == ''):
            sys.exit('Unable to detect arcmap installation path. Configure the path manually in migrate.py.')
        SetArcMapInstallDir(arcmap_dir)
    if(production_mapping_dir.strip() == ''):
        production_mapping_dir = GetProductionMappingInstallDir_AutoDetect()
        if(production_mapping_dir == ''):
            sys.exit('Unable to detect production mapping installation path. Configure the path manually in migrate.py.')
        SetProductionMappingInstallDir(production_mapping_dir)

    InitLicense()  
    
    # decode bytes to utf-8 string
    nis_paths = {}
    for k in nis_paths_orig:
        nis_paths[k.decode('utf8')] = nis_paths_orig[k]

    classes_set_migrate_all = set()
    classes_set_migrate_only_configured_series = set()
    for key in nis_paths.keys():
        cs = key.split('/')
        if len(cs) == 1:
            classes_set_migrate_all.add(cs[0])
        elif len(cs) == 2:
            classes_set_migrate_only_configured_series.add(cs[0])
        else:
            print("\t Warning: Skipping invalid configuration entry :'" + key + "'")

    # Remove conflicting configurations. Prefer the ones with 'class/series' entries over 'class' only. 
    for key in nis_paths.keys():
        cs = key.split('/')
        class_name = cs[0]
        if len(cs) == 1:
            for class_name in classes_set_migrate_only_configured_series:
                print("\t Warning: Class '" + class_name + "' has another configuration to migrate only selected series! Skipping the entry '" + key + "' :'" + nis_paths[key] + "'")
                classes_set_migrate_all.remove(class_name)
                continue

    series_path_to_id = dict()
    series_path_to_nis = dict()
    # Expand the entries with 'class' to 'class/series'
    for cs in classes_set_migrate_all:
        pl_series_table = 'PL_' + class_name + '_Series'
        for series_row in GetTablesRows(pl, pl_series_table):
            nt = RowToNamedTuple(series_row)
            series_path = class_name + "/" + nt.NAME
            series_path_to_id[series_path] = nt.GFID
            series_path_to_nis[series_path] = nis_paths[class_name]

    # For the series only configuration, consider only the series that are in the PL
    for cs in classes_set_migrate_only_configured_series:
        pl_series_table = 'PL_' + class_name + '_Series'
        for series_row in GetTablesRows(pl, pl_series_table):
            nt = RowToNamedTuple(series_row)
            series_path = class_name + "/" + nt.NAME
            if series_path in nis_paths:
                series_path_to_id[series_path] = nt.GFID
                series_path_to_nis[series_path] = nis_paths[series_path]
    # warn about the series that were not found in the PL
    for key in nis_paths:
        cs = key.split('/')
        if (len(cs) == 2) and (key not in series_path_to_id):
            print ("\t Warning: Series '" + cs[1] + "' was not found in the class '" + cs[0] + "'. Skipping the entry '" + key + "' :'" + nis_paths[key] + "'")



    # migrate the NISes
    nis_set = set(series_path_to_nis.values()) # set of NIS to migrate
    for nis in nis_set:
        print('\t Migrating NIS: ' + nis)
        nis_desc = arcpy.Describe(nis)
        if(nis_desc.workspaceType == 'RemoteDatabase'):
            print("\t\t NIS Database Connection Properties: ")
            cp = nis_desc.connectionProperties
            print("\t\t\t %-12s %s" % ("  Server:", cp.server))
            print("\t\t\t %-12s %s" % ("  Instance:", cp.instance))
            print("\t\t\t %-12s %s" % ("  User:", cp.user))
            print("\t\t\t %-12s %s" % ("  Version:", cp.version))
        MigrateNIS(nis, working_dir, chunkSize, verbose)

    if len(nis_set) == 0:
        sys.exit('\t No NIS to migrate.')

    # migrate the Product Series
    print ('\t Migrating PL: ' + pl)
    for path in series_path_to_nis.keys():
        print('\t\t Class/Series: '+ path)
        nis = series_path_to_nis[path]
        class_name, series_name = path.split('/')
        MigratePL(pl, nis, class_name, series_name, series_path_to_id[path], working_dir, agency, fids)
        

# Converts the old nis to pro by copying requred feature classes/tables and registering maritime class extension
# working_dir - dir where tempory pro nis schema is created
def MigrateNIS(arcmap_nis, working_dir, chunkSize, verbose):

    pro_nis_schema = os.path.join(working_dir, 'pro_nis_schema.gdb')

    # copy new fc/tables from temp pro nis schema to pro nis
    # WARNING: Setting 'True' will delete the ProductDefinitions and related tables.
    delete_existing = False
    if delete_existing:
        if(arcpy.Exists(os.path.join(arcmap_nis, 'ProductDefinitions'))):
            arcpy.Delete_management(os.path.join(arcmap_nis, 'ProductDefinitions'))
        if(arcpy.Exists(os.path.join(arcmap_nis, 'ProductCoverage'))):
            arcpy.Delete_management(os.path.join(arcmap_nis, 'ProductCoverage'))
        if(arcpy.Exists(os.path.join(arcmap_nis, 'ProductExports'))):
            arcpy.Delete_management(os.path.join(arcmap_nis, 'ProductExports'))
        if(arcpy.Exists(os.path.join(arcmap_nis, 'EditingProperties'))):
            arcpy.Delete_management(os.path.join(arcmap_nis, 'EditingProperties'))
        if(arcpy.Exists(os.path.join(arcmap_nis, 'S57FeatureClasses'))):
            arcpy.Delete_management(os.path.join(arcmap_nis, 'S57FeatureClasses'))

    print("\t\t Copying feature classes and tables ")
    # this copies all the associated relationship class, related tables, domains etc.
    if not arcpy.Exists(os.path.join(arcmap_nis, 'ProductDefinitions')):
        arcpy.Copy_management(os.path.join(pro_nis_schema, 'ProductDefinitions'), os.path.join(arcmap_nis, 'ProductDefinitions'))
        print("\t\t\t ProductDefinitions, ProductCoverage, ProductExports")
        if(os.path.splitext(arcmap_nis))[1] == '.sde':
            arcpy.RegisterAsVersioned_management(os.path.join(arcmap_nis, 'ProductDefinitions'))
            arcpy.RegisterAsVersioned_management(os.path.join(arcmap_nis, 'ProductCoverage'))
            arcpy.RegisterAsVersioned_management(os.path.join(arcmap_nis, 'ProductExports'))

    if not arcpy.Exists(os.path.join(arcmap_nis, 'S57FeatureClasses')):
        print("\t\t\t S57FeatureClasses" )
        arcpy.Copy_management(os.path.join(pro_nis_schema, 'S57FeatureClasses'), os.path.join(arcmap_nis, 'S57FeatureClasses'))


    if not arcpy.Exists(os.path.join(arcmap_nis, 'EditingProperties')):
        arcpy.Copy_management(os.path.join(pro_nis_schema, 'EditingProperties'), os.path.join(arcmap_nis, 'EditingProperties'))
        print("\t\t\t EditingProperties ")


    # arcpy.env.workspace = pro_nis_schema
    # all the feature classes inside the Nautical feature dataset
    # Assuming the feature class list of ENC schema is superset of all other product types
    # feature_classes = arcpy.ListFeatureClasses(feature_dataset='Nautical')

    # list of all possible feature classes inside the Nautical dataset in ENC, AML, and IENC (actually, equivalent to FCs in ENC + these 3 from AML ['IceFeaturesL', 'MilitaryFeaturesL', 'IceFeaturesP'])
    feature_classes = ['RegulatedAreasAndLimitsP', 'TracksAndRoutesP', 'TracksAndRoutesL', 'TracksAndRoutesA', 'CulturalFeaturesL', 'CulturalFeaturesA', 'CulturalFeaturesP', 'TidesAndVariationsL', 'NaturalFeaturesP', 'TidesAndVariationsA', 'IceFeaturesP', 'NaturalFeaturesA', 'DepthsA', 'NaturalFeaturesL', 'TidesAndVariationsP', 'SoundingsP', 'DangersP', 'DepthsL', 'PLTS_SpatialAttributeL', 'CoastlineP', 'OffshoreInstallationsP', 'CoastlineL', 'DangersL', 'DangersA', 'CoastlineA', 'UserDefinedFeaturesP', 'MilitaryFeaturesP', 'MilitaryFeaturesL', 'UserDefinedFeaturesA', 'AidsToNavigationP', 'UserDefinedFeaturesL', 'MilitaryFeaturesA', 'MetaDataL', 'OffshoreInstallationsL', 'OffshoreInstallationsA', 'MetaDataA', 'RegulatedAreasAndLimitsL', 'MetaDataP', 'RegulatedAreasAndLimitsA', 'IceFeaturesL', 'SeabedP', 'PortsAndServicesL', 'PortsAndServicesA', 'SeabedA', 'IceFeaturesA', 'PortsAndServicesP', 'SeabedL']

    # any additional classes/tables outside the Nautical dataset
    feature_classes.extend(['PLTS_COLLECTIONS'])
    print("\t\t Updating NOID")

    feature_classes = [fc for fc in feature_classes if arcpy.Exists(os.path.join(arcmap_nis, fc))]

    # # CopyGuidToNoid method 1
    # StartEditSession(arcmap_nis)
    # StartEditOperation(arcmap_nis)
    # start = time.time()
    # CopyGuidToNoid(arcmap_nis, feature_classes, working_dir, indent='\t\t\t')
    # StopEditOperation(arcmap_nis)
    # StopEditSession(arcmap_nis)
    # end = time.time()
    # print ("\t\t Updated NOIDs in "+ str(end - start) + "s")

    # CopyGuidToNoid method 2 (breaks into edit sessions per feature classes / details documented in the function)
    start = time.time()
    CopyGUIDToNoid2(arcmap_nis, feature_classes, working_dir, chunkSize, verbose, indent='\t\t\t')
    end = time.time()
    print ("\t\t Updated NOIDs in "+ str(end - start) + "s")

    # These tables are deprecated in pro but are required for backward compatibility
    # in arcmap for the rel manager and exporter to work properly. DO NOT DELETE these tables.
    # print("\t\t Deleting deprecated tables: ")
    # if(arcpy.Exists(os.path.join(arcmap_nis, 'PLTS_PRODUCT'))):
    #     arcpy.Delete_management(os.path.join(arcmap_nis, 'PLTS_PRODUCT'))
    #     print("\t\t\t PLTS_PRODUCT")
    # if(arcpy.Exists(os.path.join(arcmap_nis, 'PLTS_SREL'))):
    #     arcpy.Delete_management(os.path.join(arcmap_nis, 'PLTS_SREL'))
    #     print("\t\t\t PLTS_SREL")
    # if(arcpy.Exists(os.path.join(arcmap_nis, 'PLTS_NAME'))):
    #     arcpy.Delete_management(os.path.join(arcmap_nis, 'PLTS_NAME'))
    #     print("\t\t\t PLTS_NAME")

    # register maritime workspace extension
    # print("\t\t Registering maritime workspace extension ")
    # RegisterMaritimeWorkspaceExtension(OpenWorkspace(arcmap_nis))

    # register maritime class extension
    print("\t\t Registering maritime class extension ")
    # any additional feature classes/tables outside the Nautical dataset
    feature_classes_extension = copy.copy(feature_classes)
    feature_classes_extension.extend(['EditingProperties', 'ProductCoverage', 'ProductDefinitions', 'ProductExports', 'PLTS_FREL', 'PLTS_NAUTICAL_DELETES_A', 'PLTS_NAUTICAL_DELETES_L', 'PLTS_NAUTICAL_DELETES_P'])
    for fc in feature_classes_extension:
        if(arcpy.Exists(os.path.join(arcmap_nis, fc))):
            pTable = OpenTable(arcmap_nis, fc)
            RegisterMaritimeClassExtension(pTable)

    feature_classes_editor_tracking = copy.copy(feature_classes) #[os.path.join(arcmap_nis, 'Nautical', fc) for fc in feature_classes]
    feature_classes_editor_tracking.extend(['ClosingLinesL', 'PLTS_COLLECTIONS', 'ProductCoverage', 'ProductDefinitions', 'ProductExports', 'PLTS_NAUTICAL_DELETES_A', 'PLTS_NAUTICAL_DELETES_L', 'PLTS_NAUTICAL_DELETES_P'])
    print("\t\t Enabling editor tracking")
    for fc in feature_classes_editor_tracking:
        fc_fullpath = os.path.join(arcmap_nis, fc)
        if(arcpy.Exists(fc_fullpath)):
            arcpy.EnableEditorTracking_management (fc_fullpath, last_editor_field = 'NIS_EDITOR', last_edit_date_field ='NIS_LAST_MOD', record_dates_in='DATABASE_TIME')


def GetExtractionQuery(pl, class_name, series_name, product_name, inst_name, temp_dir):
    # query soln with given class name
    tables = 'PL_Solution, PL_ProductClass'
    whereClause = "PL_Solution.GFID = PL_ProductClass.ParentId and PL_ProductClass.Name = '" + class_name + "'"
    row = GetTablesRows(pl, tables, whereClause, subfields='PL_Solution.Name')[0]
    soln_name = RowToNamedTuple(row, fieldAlias = 'SOLN_NAME').SOLN_NAME
    # retrieve extraction query from the concatenated path
    entry_path = '::'.join([soln_name, class_name, series_name, product_name, inst_name])

    rows = GetTablesRows(pl, 'PL_ExtractionQuery', whereClause="EntryPath = '" + entry_path + "'")
    if len(rows) == 0:
        return None

    blob = RowToNamedTuple(rows[0]).QUERY
    extraction_query = GetPLTSSQLString(blob)
    #SELECT <TargetTable.OID> FROM <TargetTable> WHERE ((IS_CONFLATE = 1 AND PLTS_COMP_SCALE >= 25000) OR ((IS_CONFLATE = 0 OR IS_CONFLATE IS NULL) AND (PLTS_COMP_SCALE >= 25000 AND PLTS_COMP_SCALE < 45000)) ) AND NIS_PRODUCTS IN (1,2)
    # extract only the WHERE clause
    return extraction_query.split(' WHERE ')[1].strip()


# Populates ProductDefinitions, ProductCoverage, and ProductExports from arcmap PL
def MigratePL(prod_lib, nis, class_name, series_name, series_id, temp_dir, agency, fids):

    pl_product_table = 'PL_' + class_name + '_Product'
    pl_instance_table = 'PL_' + class_name + '_Instance'
    pl_aoi_table = 'PL_' + class_name + '_AOI'

    StartEditSession(nis)

    ptable_pd = OpenTable(nis, 'ProductDefinitions')
    ptable_pc = OpenTable(nis, 'ProductCoverage')
    ptable_pe = OpenTable(nis, 'ProductExports')
    preln_class_pd_pc = OpenRelationshipClass(nis, 'ProductDefinitions_ProductCoverage')
    preln_class_pd_pe = OpenRelationshipClass(nis, 'ProductDefinitions_ProductExports')

    migrated_products = set()
    for dsnm_row in GetTablesRows(nis, 'ProductDefinitions', subfields="DSNM"):
        nt = RowToNamedTuple(dsnm_row)
        dsnm = nt.DSNM.split('.')[0]
        migrated_products.add(dsnm)

    for ptrow in GetTablesRows(prod_lib, pl_product_table, whereClause="ParentId = '" + series_id + "'"):
        nt = RowToNamedTuple(ptrow)
        product_id = nt.GFID
        product_name = nt.NAME

        if(product_name in migrated_products):
            print("\t\t\t Product: " + product_name + " is already migrated. Skipping.")
            continue

        print("\t\t\t Product: " + product_name)

        # Assuming only one instance in the series, or always using the 1st one.
        inst_name = ''
        inst_id = ''
        for inst_row in GetTablesRows(prod_lib, pl_instance_table, whereClause="ParentId = '" + product_id + "'"):
            inst_name = RowToNamedTuple(inst_row).NAME
            inst_id = RowToNamedTuple(inst_row).GFID
            break
        
        extraction_query = GetExtractionQuery(prod_lib, class_name, series_name, product_name, inst_name, temp_dir)
        if extraction_query is None:
            print("\t\t\t\t Warning: Unable to find extraction query for the product {}. Skipping.".format(product_name))
            continue

        # populate ProductDefinition table
        # ------------------------------------------------------------------------------------------
        maxEDTN = 0
        UPDN = 0
        # First check if there are any product with creation type != New_Cell.
        # New_Cell may have Edtn > 1 so make sure to use EDTN of non-New_Cell first.
        metadata_rows = GetTablesRows(prod_lib, 'NAUTICAL_METADATA', whereClause="PRODUCT_ID = '" + product_id + "' and CREATION_TYPE <> 3") # 3 = New_Cell
        if len(metadata_rows) > 0:
            mds = []
            for mr in metadata_rows:
                nt = RowToNamedTuple(mr)
                # CREATION_TYPES (1:Import, 2:New_Dataset, 3:New_Cell, 4:New_Edition, 5:Update, 6:Reissue 7:Cancel)
                is_reissue = 1 if nt.CREATION_TYPE == 6 else 0
                is_cancel = 1 if nt.CREATION_TYPE == 7 else 0
                mds.append([is_cancel, nt.EDTN, nt.UPDN, is_reissue, nt.NAUTICAL_METADATA])
            # sort with the keys in this order: cancel, edtn, updn, reissue 
            # (cancel is always considered latest, followed by highest edtn, and updn,
            #  for same edtn and updn, reissue breaks the tie i.e. reissue is considered the latest)
            mx = sorted(mds, key=operator.itemgetter(0, 1, 2, 3), reverse=True)[0]
            maxEDTN = mx[1]
            UPDN = mx[2]
            metadata_xml_element = mx[4]
        else:
            nt = RowToNamedTuple(GetTablesRows(prod_lib, 'NAUTICAL_METADATA', whereClause="PRODUCT_ID = '" + product_id + "'")[0])
            maxEDTN = nt.EDTN
            UPDN = nt.UPDN
            metadata_xml_element = nt.NAUTICAL_METADATA

        # CREATION_TYPES (1:Import, 2:New_Dataset, 3:New_Cell, 4:New_Edition, 5:Update, 6:Reissue 7:Cancel)
        ct_code = metadata_xml_element.find('CREATION_TYPE').text
        if ct_code == '2':
            export_type = 'NewDataset'
        elif ct_code == '4':
            export_type = 'NewEdition'
        elif ct_code == '5':
            export_type = 'Update'
        elif ct_code == '6':
            export_type = 'Reissue'
        elif ct_code == '7':
            export_type = 'Cancel'
        else: # '1', '3' implies product has not been exported yet, so null
            export_type = None

        cscl =  metadata_xml_element.find('INSTANCE/CSCL').text
        values = {
            'CSCL' : cscl,
            'DSNM' : metadata_xml_element.find('INSTANCE/DSNM').text, 
            'EDTN' : maxEDTN, 
            'UPDN' : UPDN,
            'ExportType' : export_type,
            # PROF (1:EN, 2:ER) is empty in th metadata xml, so use export_type(CREATION_TYPE),
            'PROF' : 2 if (export_type == 'Update' or export_type == 'Cancel') else 1,
            'INTU' : metadata_xml_element.find('INTU').text, 
            'UADT' : metadata_xml_element.find('UADT').text, 
            'ISDT' : metadata_xml_element.find('ISDT').text, 
            'STED' : metadata_xml_element.find('STED').text, 
            'PRSP' : metadata_xml_element.find('PRSP').text, 
            'PSDN' : metadata_xml_element.find('PSDN').text, 
            'PRED' : metadata_xml_element.find('PRED').text, 
            'AGEN' : metadata_xml_element.find('AGEN').text, 
            'COMT' : metadata_xml_element.find('COMT').text, 
            'AALL' : metadata_xml_element.find('AALL').text, 
            'NALL' : metadata_xml_element.find('NALL').text, 
            'HDAT' : metadata_xml_element.find('INSTANCE/HDAT').text, 
            'VDAT' : metadata_xml_element.find('INSTANCE/VDAT').text, 
            'SDAT' : metadata_xml_element.find('INSTANCE/SDAT').text, 
            'DUNI' : metadata_xml_element.find('INSTANCE/DUNI').text, 
            'HUNI' : metadata_xml_element.find('INSTANCE/HUNI').text, 
            'PUNI' : metadata_xml_element.find('INSTANCE/PUNI').text, 
            'COUN' : metadata_xml_element.find('INSTANCE/COUN').text, 
            'COMF' : metadata_xml_element.find('INSTANCE/COMF').text, 
            'SOMF' : metadata_xml_element.find('INSTANCE/SOMF').text, 
            'DSPM_COMT' : metadata_xml_element.find('INSTANCE/COMT').text,
            'Series': series_name,
            'WhereClause' : extraction_query
        }

        if values['VDAT'] is None:
            values['VDAT'] = 1
            print("\t\t\t Warning: Mandatory ProductDefinitions field VDAT was empty. Setting value to '1. Mean low water springs'")
        if values['SDAT'] is None:
            values['SDAT'] = 1
            print("\t\t\t Warning: Mandatory ProductDefinitions field SDAT was empty. Setting value to '1. Mean low water springs'")

        StartEditOperation(nis)
        pd_row = InsertIntoTable(ptable_pd, values)
        StopEditOperation(nis)

        # populate ProductCoverage table and add relationship ProductDefinition_ProductCoverage
        # ------------------------------------------------------------------------------------------
        aoi_rows = GetTablesRows(prod_lib, pl_aoi_table, whereClause="ParentId = '" + inst_id + "'")
        for aoirow in aoi_rows:
            nt = RowToNamedTuple(aoirow)
            values = {
                'DSNM': product_name,
                'CATCOV': 1 if nt.USEDATA == -1 else 2, # Code_CKB_Bool to NAUTICAL_CATCOV
                'SHAPE': nt.SHAPE,
                'LNAM': agency + GenerateFIDN() + fids,
                'PLTS_COMP_SCALE': cscl
            }
            StartEditOperation(nis)
            pc_row = InsertIntoTable(ptable_pc, values)
            preln_class_pd_pc.CreateRelationship(pd_row, pc_row)
            StopEditOperation(nis)


        # populate ProductExports table and add relationship ProductDefinition_ProductExports
        # ------------------------------------------------------------------------------------------

        # PL_Relationship.OriginId points to a Product id and PL_Relationsihp.DestinationId points to PL_File.GFID.
        # So PL_Relationship links product id to one or more file (mxd, zip etc) in  PL_File.

        FileDataTuple = namedtuple('FileDataTuple', 'EDTN UPDN EXPORTTYPE UADT ISDT EXPORTDATA')
        file_data_tuples = []

        # Find zip file ids related to the product.
        subfields = 'PL_File.GFID'
        tables = 'PL_Relationship, PL_File'
        whereClause = "PL_Relationship.OriginId ='" + product_id + "' and \
            PL_Relationship.DestinationId = PL_File.GFID and \
            (PL_File.Name = '"+ product_name + ".zip' or PL_File.Name = '" + product_name + "_Cancel.zip')"
        file_rows = GetTablesRows(prod_lib, tables, whereClause, subfields)

        for fr in file_rows:
            nt = RowToNamedTuple(fr, fieldAlias='pl_file_gfid')
            tables = 'PL_FileVersion'
            subfields = 'GFID, Major, Minor, Revision'
            whereClause = "PL_FileVersion.FileId = '" + nt.pl_file_gfid + "'"
            fileversion_rows = GetTablesRows(prod_lib, tables, whereClause, subfields)
            for fvr in fileversion_rows:
                nt1 = RowToNamedTuple(fvr)
                major = nt1.MAJOR
                minor = nt1.MINOR
                revision = nt1.REVISION
                tables = 'PL_FileData'
                subfields = 'FileData'
                whereClause = "PL_FileData.FileVersionId = '" + nt1.GFID + "'"
                filedata_rows = GetTablesRows(prod_lib, tables, whereClause, subfields)
                for fdr in filedata_rows:
                    nt2 = RowToNamedTuple(fdr)
                    filedata = nt2.FILEDATA

                    for mdr in metadata_rows:
                        nt3 = RowToNamedTuple(mdr)

                        # CREATION_TYPES in NAUTICAL_METADATA table:
                        # 1:Import, 2:New_Dataset, 3:New_Cell, 4:New_Edition, 5:Update, 6:Reissue 7:Cancel)
                        if nt3.CREATION_TYPE == 2:
                            export_type = 'NewDataset'
                        elif nt3.CREATION_TYPE == 4:
                            export_type = 'NewEdition'
                        elif nt3.CREATION_TYPE == 5:
                            export_type = 'Update'
                        elif nt3.CREATION_TYPE == 6:
                            export_type = 'Reissue'
                        elif nt3.CREATION_TYPE == 7:
                            export_type = 'Cancel'
                        else: # 1, 3 implies product has not been exported yet, so null (this branch should not be hit for the exported data.)
                            export_type = None

                        revision_from_metadata_table = 1 if export_type == 'Reissue' else 0
                        
                        # The following if condition is same as join of PL_FileVersion and NAUTICAL_METADATA tables with:
                        # PL_FileVersion.Major = NAUTICAL_METADATA.EDTN and 
                        # PL_FileVersion.Minor = NAUTICAL_METADATA.UPDN and
                        # PL_FileVersion.Revision = NAUTICAL_METADATA.CREATION_TYPE of value 6
                        if(nt3.EDTN == major and nt3.UPDN == minor and revision_from_metadata_table == revision):
                            uadt = nt3.NAUTICAL_METADATA.find('UADT').text
                            isdt = nt3.NAUTICAL_METADATA.find('ISDT').text
                            file_data_tuples.append(FileDataTuple(major, minor, export_type, uadt, isdt, filedata))

        for fdt in file_data_tuples:
            msg = '\t\t\t\t  EDTN: ' + str(fdt.EDTN) + ' UPDN: ' + str(fdt.UPDN) + ' (' + fdt.EXPORTTYPE + ')'
            print(msg)

            values = {
                'DSNM': product_name,
                'EDTN': fdt.EDTN,
                'UPDN': fdt.UPDN,
                'ExportType' : fdt.EXPORTTYPE,
                'PROF' : 2 if (fdt.EXPORTTYPE == 'Update' or fdt.EXPORTTYPE == 'Cancel') else 1,
                'UADT' : fdt.UADT, 
                'ISDT' : fdt.ISDT, 
                'EXPORTDATA': fdt.EXPORTDATA
            }
            StartEditOperation(nis)
            pe_row = InsertIntoTable(ptable_pe, values)
            preln_class_pd_pe.CreateRelationship(pd_row, pe_row)
            StopEditOperation(nis)

        #-------------------------------------------------------------------------------------------
        # This query is same as the above queries for populating ProductExports table, however 
        # it times out on some Oracle connection probably because of the number of joins
        #  
        # # subfields = 'NAUTICAL_METADATA.EDTN, NAUTICAL_METADATA.UPDN, NAUTICAL_METADATA.CREATION_TYPE, NAUTICAL_METADATA.NAUTICAL_METADATA, PL_FileVersion.Revision, PL_FileData.FileData'
        # # tables = 'PL_Relationship, PL_File, PL_FileVersion, NAUTICAL_METADATA, PL_FileData'
        # # whereClause = "PL_Relationship.OriginId ='" + product_id + "' and \
        # #             PL_Relationship.DestinationId = PL_File.GFID and \
        # #             (PL_File.Name = '"+ product_name + ".zip' or PL_File.Name = '" + product_name + "_Cancel.zip') and \
        # #             PL_FileVersion.FileId = PL_File.GFID and \
        # #             PL_FileData.FileVersionId = PL_FileVersion.GFID and \
        # #             NAUTICAL_METADATA.PRODUCT_ID = '" + product_id + "' and \
        # #             NAUTICAL_METADATA.CREATION_TYPE <> 3 and \
        # #             PL_FileVersion.Major = NAUTICAL_METADATA.EDTN and \
        # #             PL_FileVersion.Minor = NAUTICAL_METADATA.UPDN"
        # # filedata_rows = GetTablesRows(prod_lib, tables, whereClause, subfields)
        #-------------------------------------------------------------------------------------------

    StopEditSession(nis)



